#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import itertools
import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.contrib.pynumero.dependencies import (
    numpy as np, numpy_available, scipy, scipy_available
)

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.common.dependencies.scipy import sparse as sps

from pyomo.contrib.pynumero.asl import AmplInterface
if not AmplInterface.available():
    raise unittest.SkipTest(
        "Pynumero needs the ASL extension to run cyipopt tests")

from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import (
    cyipopt_available,
)
from pyomo.contrib.pynumero.interfaces.external_pyomo_model import (
    ExternalPyomoModel,
    get_hessian_of_constraint,
)

if not pyo.SolverFactory("ipopt").available():
    raise unittest.SkipTest(
        "Need IPOPT to run ExternalPyomoModel tests"
        )


class SimpleModel1(object):

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=2.0)
        m.y = pyo.Var(initialize=2.0)
        m.residual_eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        m.external_eqn = pyo.Constraint(expr=m.x*m.y == 0.2)
        # The "external function constraint" exposed by the ExternalPyomoModel
        # will look like: x**2 + 0.04/x**2 - 1 == 0
        return m

    def evaluate_external_variables(self, x):
        # y(x)
        return 0.2/x

    def evaluate_external_jacobian(self, x):
        # dydx
        return -0.2/(x**2)

    def evaluate_external_hessian(self, x):
        # d2ydx2
        return 0.4/(x**3)

    def evaluate_residual(self, x):
        return x**2 + 0.04/x**2 - 1

    def evaluate_jacobian(self, x):
        return 2*x - 0.08/x**3

    def evaluate_hessian(self, x):
        return 2 + 0.24/x**4


class SimpleModel2(object):
    """
    The purpose of this model is to exercise each term in the computation
    of the d2ydx2 Hessian.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=2.0)
        m.y = pyo.Var(initialize=2.0)
        m.residual_eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        m.external_eqn = pyo.Constraint(expr=(m.x**3)*(m.y**3) == 0.2)
        return m

    def evaluate_external_variables(self, x):
        return 0.2**(1/3)/x

    def evaluate_external_jacobian(self, x):
        return -(0.2**(1/3))/(x**2)

    def evaluate_external_hessian(self, x):
        return 2*0.2**(1/3)/(x**3)

    def evaluate_residual(self, x):
        return x**2 + 0.2**(2/3)/x**2 - 1

    def evaluate_jacobian(self, x):
        return 2*x - 2*0.2**(2/3)/x**3

    def evaluate_hessian(self, x):
        return 2 + 6*0.2**(2/3)/x**4


class SimpleModel2by2_1(object):
    """
    The purpose of this model is to test second derivative computation
    when the external model is nonlinear only in x. This exercises
    the first term in the second derivative implicit function theorem.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([0, 1], initialize=2.0)
        m.y = pyo.Var([0, 1], initialize=2.0)

        def residual_eqn_rule(m, i):
            # These equations are chosen to exercise every term in the
            # equality hessian calculation, i.e. to have nonlinearities
            # in x, y, and xy.
            if i == 0:
                return m.x[0]**2 + m.x[0]*m.y[0] + m.y[0]**2 == 1.0
            elif i == 1:
                return m.x[1]**2 + m.x[1]*m.y[1] == 2.0

        m.residual_eqn = pyo.Constraint([0, 1], rule=residual_eqn_rule)

        def external_eqn_rule(m, i):
            if i == 0:
                return m.y[0] + m.y[1] + m.x[0]*m.x[1] + m.x[0]**2 == 1.0
            elif i == 1:
                return m.y[0] + 2.0*m.x[0]*m.x[1] + m.x[1]**2 == 2.0

        m.external_eqn = pyo.Constraint([0, 1], rule=external_eqn_rule)

        return m

    # The following three methods are evaluation and derivatives of
    # the equality constraints exposed by the ExternalPyomoModel
    def evaluate_residual(self, x):
        f0 = (
                x[0]**2 + 2*x[0] - 2*x[0]**2*x[1] - x[1]**2*x[0] + 4
                - 8*x[0]*x[1] - 4*x[1]**2 + 4*x[0]**2*x[1]**2
                + 4*x[0]*x[1]**3 + x[1]**4 - 1.0
                )
        f1 = x[1]**2 - x[1] + x[0]*x[1]**2 + x[1]**3 - x[0]**2*x[1] - 2.0
        return (f0, f1)

    def evaluate_jacobian(self, x):
        df0dx0 = (
                2*x[0] + 2 - 4*x[0]*x[1] - x[1]**2
                - 8*x[1] + 8*x[0]*x[1]**2 + 4*x[1]**3
                )
        df0dx1 = (
                -2*x[0]**2 - 2*x[0]*x[1] - 8*x[0] - 8*x[1]
                + 8*x[0]**2*x[1] + 12*x[0]*x[1]**2 + 4*x[1]**3
                )
        df1dx0 = x[1]**2 - 2*x[0]*x[1]
        df1dx1 = 2*x[1] - 1+ 2*x[0]*x[1] - x[0]**2 + 3*x[1]**2
        return np.array([[df0dx0, df0dx1], [df1dx0, df1dx1]])

    def evaluate_hessian(self, x):
        df0dx0dx0 = 2 - 4*x[1] + 8*x[1]**2
        df0dx0dx1 = -4*x[0] - 2*x[1] - 8 + 16*x[0]*x[1] + 12*x[1]**2
        df0dx1dx1 = -2*x[0] - 8 + 8*x[0]**2 + 24*x[0]*x[1] + 12*x[1]**2

        df1dx0dx0 = -2*x[1]
        df1dx0dx1 = 2*x[1] - 2*x[0]
        df1dx1dx1 = 2 + 2*x[0] + 6*x[1]
        d2f0 = np.array([[df0dx0dx0, df0dx0dx1], [df0dx0dx1, df0dx1dx1]])
        d2f1 = np.array([[df1dx0dx0, df1dx0dx1], [df1dx0dx1, df1dx1dx1]])
        return [d2f0, d2f1]

    # The following three methods are evaluation and derivatives of
    # the external function "hidden by" the ExternalPyomoModel
    def evaluate_external_variables(self, x):
        y0 = 2.0 - 2.0*x[0]*x[1] - x[1]**2
        y1 = 1.0 - y0 - x[0]*x[1] - x[0]**2
        return (y0, y1)

    def evaluate_external_jacobian(self, x):
        dy0dx0 = -2.0*x[1]
        dy0dx1 = -2.0*x[0] - 2.0*x[1]
        dy1dx0 = -dy0dx0 - x[1] - 2.0*x[0]
        dy1dx1 = -dy0dx1 - x[0]
        return np.array([[dy0dx0, dy0dx1], [dy1dx0, dy1dx1]])

    def evaluate_external_hessian(self, x):
        dy0dx0dx0 = 0.0
        dy0dx0dx1 = -2.0
        dy0dx1dx1 = -2.0

        dy1dx0dx0 = -dy0dx0dx0 - 2.0
        dy1dx0dx1 = -dy0dx0dx1 - 1.0
        dy1dx1dx1 = -dy0dx1dx1

        dy0dxdx = np.array([[dy0dx0dx0, dy0dx0dx1], [dy0dx0dx1, dy0dx1dx1]])
        dy1dxdx = np.array([[dy1dx0dx0, dy1dx0dx1], [dy1dx0dx1, dy1dx1dx1]])
        return [dy0dxdx, dy1dxdx]


class Model2by2(object):
    """
    The purpose of this model is to test d2ydx2 Hessian computation when
    transposes result in a nontrivial modification of Hessian/Jacobian
    matrices.
    """

    def make_model(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([0, 1], initialize=2.0)
        m.y = pyo.Var([0, 1], initialize=2.0)

        m.residual_eqn = pyo.Constraint(expr=
                m.x[0]**2 + m.x[1]**2 + m.y[0]**2 + m.y[1]**2 == 1.0
                )

        def external_eqn_rule(m, i):
            if i == 0:
                return (m.x[0]**2) * m.y[0] * (m.x[1]**0.5) * m.y[1] - 0.1 == 0
            elif i == 1:
                return m.x[0] * m.y[0] * m.x[1] - 0.2 == 0

        m.external_eqn = pyo.Constraint([0, 1], rule=external_eqn_rule)

        return m

    def evaluate_external_variables(self, x):
        y0 = 0.2/(x[0]*x[1])
        y1 = 0.1/(x[0]**2 * y0 * x[1]**0.5)
        return [y0, y1]

    def evaluate_external_jacobian(self, x):
        dy0dx0 = -0.2/(x[0]**2 * x[1])
        dy0dx1 = -0.2/(x[0] * x[1]**2)
        dy1dx0 = -0.5*x[1]**0.5/x[0]**2
        dy1dx1 = 0.25/(x[0] * x[1]**0.5)
        return np.array([[dy0dx0, dy0dx1], [dy1dx0, dy1dx1]])

    def evaluate_external_hessian(self, x):
        d2y0dx0dx0 = 0.4/(x[0]**3 * x[1])
        d2y0dx0dx1 = 0.2/(x[0]**2 * x[1]**2)
        d2y0dx1dx1 = 0.4/(x[0] * x[1]**3)

        d2y1dx0dx0 = x[1]**0.5/(x[0]**3)
        d2y1dx0dx1 = -0.25/(x[0]**2 * x[1]**0.5)
        d2y1dx1dx1 = -0.125/(x[0] * x[1]**1.5)

        d2y0dxdx = np.array([
            [d2y0dx0dx0, d2y0dx0dx1],
            [d2y0dx0dx1, d2y0dx1dx1],
            ])
        d2y1dxdx = np.array([
            [d2y1dx0dx0, d2y1dx0dx1],
            [d2y1dx0dx1, d2y1dx1dx1],
            ])
        return [d2y0dxdx, d2y1dxdx]


class TestGetHessianOfConstraint(unittest.TestCase):

    def test_simple_model_1(self):
        model = SimpleModel1()
        m = model.make_model()
        m.x.set_value(2.0)
        m.y.set_value(2.0)

        con = m.residual_eqn
        expected_hess = np.array([[2.0, 0.0], [0.0, 2.0]])
        hess = get_hessian_of_constraint(con)
        self.assertTrue(np.all(expected_hess == hess.toarray()))

        expected_hess = np.array([[2.0]])
        hess = get_hessian_of_constraint(con, [m.x])
        self.assertTrue(np.all(expected_hess == hess.toarray()))

        con = m.external_eqn
        expected_hess = np.array([[0.0, 1.0], [1.0, 0.0]])
        hess = get_hessian_of_constraint(con)
        self.assertTrue(np.all(expected_hess == hess.toarray()))

    def test_polynomial(self):
        m = pyo.ConcreteModel()

        n_x = 3
        x1 = 1.1
        x2 = 1.2
        x3 = 1.3
        m.x = pyo.Var(range(1, n_x+1), initialize={1: x1, 2: x2, 3: x3})
        m.eqn = pyo.Constraint(expr=
                5*(m.x[1]**5) +                        # T1
                5*(m.x[1]**4)*(m.x[2]) +               # T2
                5*(m.x[1]**3)*(m.x[2])*(m.x[3]) +      # T3
                5*(m.x[1])*(m.x[2]**2)*(m.x[3]**2) +   # T4
                4*(m.x[1]**2)*(m.x[2])*(m.x[3]) +      # T5
                4*(m.x[2]**2)*(m.x[3]**2) +            # T6
                4*(m.x[3]**4) +                        # T7
                3*(m.x[1])*(m.x[2])*(m.x[3]) +         # T8
                3*(m.x[2]**3) +                        # T9
                3*(m.x[2]**2)*(m.x[3]) +               # T10
                2*(m.x[1])*(m.x[2]) +                  # T11
                2*(m.x[2])*(m.x[3])                    # T12
                == 0
                )

        rcd = []
        rcd.append((0, 0, (
            # wrt x1, x1
            5*5*4*x1**3 +    # T1
            5*4*3*x1**2*x2 + # T2
            5*3*2*x1*x2*x3 + # T3
            4*2*1*x2*x3      # T5
            )))
        rcd.append((1, 1, (
            # wrt x2, x2
            5*x1*2*x3**2 +   # T4
            4*2*x3**2 +      # T6
            3*3*2*x2 +       # T9
            3*2*x3           # T10
            )))
        rcd.append((2, 2, (
            # wrt x3, x3
            5*x1*x2**2*2 +   # T4
            4*x2**2*2 +      # T6
            4*4*3*x3**2      # T7
            )))
        rcd.append((1, 0, (
            # wrt x2, x1
            5*4*x1**3 +      # T2
            5*3*x1**2*x3 +   # T3
            5*2*x2*x3**2 +   # T4
            4*2*x1*x3 +      # T5
            3*x3 +           # T8
            2                # T11
            )))
        rcd.append((2, 0, (
            # wrt x3, x1
            5*3*x1**2*x2 +   # T3
            5*x2**2*2*x3 +   # T4
            4*2*x1*x2 +      # T5
            3*x2             # T8
            )))
        rcd.append((2, 1, (
            # wrt x3, x2
            5*x1**3 +        # T3
            5*x1*2*x2*2*x3 + # T4
            4*x1**2 +        # T5
            4*2*x2*2*x3 +    # T6
            3*x1 +           # T8
            3*2*x2 +         # T10
            2                # T12
            )))

        row = [r for r, _, _ in rcd]
        col = [c for _, c, _ in rcd]
        data = [d for _, _, d in rcd]
        expected_hess = sps.coo_matrix((data, (row, col)), shape=(n_x, n_x))
        expected_hess_array = expected_hess.toarray()
        expected_hess_array = (
                expected_hess_array
                + np.transpose(expected_hess_array)
                - np.diag(np.diagonal(expected_hess_array))
                )
        hess = get_hessian_of_constraint(m.eqn, list(m.x.values()))
        hess_array = hess.toarray()
        np.testing.assert_allclose(expected_hess_array, hess_array, rtol=1e-8)

    def test_unused_variable(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.0)
        m.y = pyo.Var(initialize=1.0)
        m.z = pyo.Var(initialize=1.0)
        m.eqn = pyo.Constraint(expr=m.x**2 + m.y**2 == 1.0)
        variables = [m.x, m.y, m.z]
        expected_hess = np.array([[2, 0, 0], [0, 2, 0], [0, 0, 0]])
        hess = get_hessian_of_constraint(m.eqn, variables).toarray()
        np.testing.assert_allclose(hess, expected_hess, rtol=1e-8)

    def test_explicit_zeros(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1.0)
        m.y = pyo.Var(initialize=0.0)
        m.eqn = pyo.Constraint(expr=m.x**2 + m.y**3 == 1.0)
        variables = [m.x, m.y]

        row = np.array([0, 1])
        col = np.array([0, 1])
        data = np.array([2.0, 0.0])
        expected_hess = sps.coo_matrix(
                (data, (row, col)), shape=(2,2)
                )
        hess = get_hessian_of_constraint(m.eqn, variables)
        np.testing.assert_allclose(hess.row, row, atol=0)
        np.testing.assert_allclose(hess.col, col, atol=0)
        np.testing.assert_allclose(hess.data, data, rtol=1e-8)


class TestExternalPyomoModel(unittest.TestCase):

    def test_evaluate_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            self.assertAlmostEqual(
                    resid[0],
                    model.evaluate_residual(x[0]),
                    delta=1e-8,
                    )

    def test_jacobian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            self.assertAlmostEqual(
                    jac.toarray()[0][0],
                    model.evaluate_jacobian(x[0]),
                    delta=1e-8,
                    )

    def test_hessian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            self.assertAlmostEqual(
                    hess[0][0, 0],
                    model.evaluate_hessian(x[0]),
                    delta=1e-8,
                    )

    def test_external_hessian_SimpleModel1(self):
        model = SimpleModel1()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x[0])
            self.assertAlmostEqual(
                    hess[0][0,0],
                    expected_hess,
                    delta=1e-8,
                    )

    def test_evaluate_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            self.assertAlmostEqual(
                    resid[0],
                    model.evaluate_residual(x[0]),
                    delta=1e-8,
                    )

    def test_jacobian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            # evaluate_jacobian_equality_constraints involves an LU
            # factorization and repeated back-solve. SciPy returns a
            # dense matrix from this operation. I am not sure if I should
            # cast it to a sparse matrix. For now it is dense...
            self.assertAlmostEqual(
                    jac.toarray()[0][0],
                    model.evaluate_jacobian(x[0]),
                    delta=1e-7,
                    )

    def test_hessian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            self.assertAlmostEqual(
                    hess[0][0, 0],
                    model.evaluate_hessian(x[0]),
                    delta=1e-7,
                    )

    def test_external_jacobian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x[0])
            self.assertAlmostEqual(
                    jac[0,0],
                    expected_jac,
                    delta=1e-8,
                    )

    def test_external_hessian_SimpleModel2(self):
        model = SimpleModel2()
        m = model.make_model()
        x_init_list = [
                [-5.0], [-4.0], [-3.0], [-1.5], [0.5], [1.0], [2.0], [3.5]
                ]
        external_model = ExternalPyomoModel(
                [m.x], [m.y], [m.residual_eqn], [m.external_eqn],
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x[0])
            self.assertAlmostEqual(
                    hess[0][0,0],
                    expected_hess,
                    delta=1e-7,
                    )

    def test_external_jacobian_Model2by2(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x)
            np.testing.assert_allclose(jac, expected_jac, rtol=1e-8)

    def test_external_hessian_Model2by2(self):
        model = Model2by2()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [0.5, 1.0, 1.5, 2.5, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x)
            for matrix1, matrix2 in zip(hess, expected_hess):
                matrix2 = np.matrix(matrix2)
                np.testing.assert_allclose(matrix1, matrix2, rtol=1e-8)

    def test_external_jacobian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_external_variables()
            expected_jac = model.evaluate_external_jacobian(x)
            np.testing.assert_allclose(jac, expected_jac, rtol=1e-8)

    def test_external_hessian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessian_external_variables()
            expected_hess = model.evaluate_external_hessian(x)
            for matrix1, matrix2 in zip(hess, expected_hess):
                np.testing.assert_allclose(matrix1, matrix2, rtol=1e-8)

    def test_evaluate_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            resid = external_model.evaluate_equality_constraints()
            expected_resid = model.evaluate_residual(x)
            np.testing.assert_allclose(resid, expected_resid, rtol=1e-8)

    def test_jacobian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            jac = external_model.evaluate_jacobian_equality_constraints()
            expected_jac = model.evaluate_jacobian(x)
            np.testing.assert_allclose(jac.toarray(), expected_jac, rtol=1e-8)

    def test_hessian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            hess = external_model.evaluate_hessians_of_residuals()
            expected_hess = model.evaluate_hessian(x)
            np.testing.assert_allclose(hess, expected_hess, rtol=1e-8)

    def test_evaluate_hessian_lagrangian_SimpleModel2x2_1(self):
        model = SimpleModel2by2_1()
        m = model.make_model()
        m.x[0].set_value(1.0)
        m.x[1].set_value(2.0)
        m.y[0].set_value(3.0)
        m.y[1].set_value(4.0)
        x0_init_list = [-5.0, -3.0, 0.5, 1.0, 2.5]
        x1_init_list = [-4.5, -2.3, 0.0, 1.0, 4.1]
        x_init_list = list(itertools.product(x0_init_list, x1_init_list))
        external_model = ExternalPyomoModel(
                list(m.x.values()),
                list(m.y.values()),
                list(m.residual_eqn.values()),
                list(m.external_eqn.values()),
                )

        for x in x_init_list:
            external_model.set_input_values(x)
            external_model.set_equality_constraint_multipliers([1.0, 1.0])
            hess_lag = external_model.evaluate_hessian_equality_constraints()
            hess_lag = hess_lag.toarray()
            expected_hess = model.evaluate_hessian(x)
            expected_hess_lag = np.tril(expected_hess[0] + expected_hess[1])
            np.testing.assert_allclose(hess_lag, expected_hess_lag, rtol=1e-8)


if __name__ == '__main__':
    unittest.main()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
from pyomo.core.base import Block, Reference
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.base.block import SubclassOf
from pyomo.core.base.set import SetProduct
from pyomo.core.base.indexed_component import (
        UnindexedComponent_set,
        normalize_index,
        )
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from collections import OrderedDict


def get_slice_for_set(s):
    """
    Get the slice of the proper dimension for a set.
    """
    if s.dimen != 0:
        if not normalize_index.flatten:
            return slice(None)
        else:
            if s.dimen is not None:
                # We will arrive here and fail for sets of dimension
                # UnknownSetDimen.
                return (slice(None),)*s.dimen
            else:
                return (Ellipsis,)
    else:
        # Case for e.g. UnindexedComponent_set
        return None

class _NotAnIndex(object):
    """ 
    `None` is a valid index, so we use a dummy class to 
    denote a slot that needs to get filled with indices
    from our product.
    """
    pass

def _fill_indices(filled_index, index):
    """
    `filled_index` is a list with some entries `_NotAnIndex`.
    We fill those entries with values from `index`, a tuple.
    """
    j = 0
    for i, val in enumerate(filled_index):
        if val is _NotAnIndex:
            filled_index[i] = index[j]
            # `index` is always a tuple, so this is valid
            j += 1
    # Make sure `filled_index` had the same number of vacancies
    # as the original SetProduct has factors. Not _strictly_ necessary.
    assert j == len(index)
    filled_index = tuple(filled_index)

    if len(filled_index) == 1:
        return filled_index[0]
    else:
        return filled_index

def _fill_indices_from_product(partial_index_list, product):
    """ 
    `partial_index_list` is a list of indices, each corresponding to a
    set. If an entry in `partial_index_list` is `_NotAnIndex`, that
    slot will get filled in by an entry from `product`.

    `product` is a `SetProduct` with as many "factors" as there are
    missing indices in `partial_index_list`.
    """
    # We will manipulate `normalize_index.flatten`.
    # Store its original value so we can reset it when we're done.
    _normalize_index_flatten = normalize_index.flatten
    try:
        normalize_index.flatten = False
        for index in product:
            # Since `normalize_index.flatten` is False, `index` is a
            # scalar or (tuple of (scalars or tuples)). Conveniently,
            # each entry in the tuple belongs to a single factor set.
            # I do not have to worry about having a product-of-products
            # here because I created the product from "unfactorable sets"

            # We need to generate a new index for every entry of `product`,
            # and want to reuse `partial_index_list` as a starting point,
            # so we copy it here.
            filled_index = list(partial_index_list)

            normalize_index.flatten = _normalize_index_flatten

            # filled_index can now be used in the user's intended way
            # This determines how we will access the component's data
            # with our new index, which is currently _completely_ unflattened
            # (i.e. a tuple-of-tuples, no further nesting).
            #
            # This will not work well if the user's component is indexed by
            # a nested product of sets AND normalize_index.flatten is False.
            # In this case we may try to access a component's data with
            # >>> comp[(1,'a',1)]
            # (each coordinate belongs to its own set) when it expects
            # >>> comp[((1,'a'),1)]
            # because `comp` was created with two set arguments, the first
            # of which was already a product.

            yield _fill_indices(filled_index, index)

            normalize_index.flatten = False
            # Want to get the unflattened factors when we advance the
            # iterator of `product`
    finally:
        # Reset `normalize_index.flatten`
        normalize_index.flatten = _normalize_index_flatten

def generate_sliced_components(b, index_stack, slice_, sets, ctype, index_map):
    """
    Recursively generate sliced components of a block and its subblocks, along
    with the sets that were sliced for each component.

    `b` is a _BlockData object.

    `index_stack` is a list of indices "above" `b` in the
    hierarchy. Note that `b` is a data object, so any index
    of its parent component should be included in the stack.

    `slice_` is the slice generated so far. Our goal here is to
    yield extensions to `slice_` at this level of the hierarchy.

    `sets` is a ComponentSet of Pyomo sets that should be sliced.

    `ctype` is the type we are looking for.

    `index_map` is potentially a map from each set in `sets` to a 
    "representative index" to use when descending into subblocks.
    """
    for c in b.component_objects(ctype, descend_into=False):
        subsets = list(c.index_set().subsets())
        temp_idx = [get_slice_for_set(s) if s in sets else _NotAnIndex
                for s in subsets]
        new_sets = [s for s in subsets if s in sets]
        other_sets = [s for s in subsets if s not in sets]
        sliced_sets = index_stack + new_sets

        # We have extended our "index stack;" now we must extend
        # our slice.

        if other_sets and c.is_indexed():
            # We need to iterate over sets that aren't sliced
            # `c.is_indexed()` covers the case when UnindexedComponent_set
            # is in `other_sets`.
            cross_prod = other_sets[0].cross(*other_sets[1:])
            # The original implementation was to pick an arbitrary index
            # from the "flattened sets" and slice all the other indices.
            #
            # This implementation avoids issues about selecting an arbitrary
            # index, but requires duplicating some of the slice-iter logic below

            for new_index in _fill_indices_from_product(temp_idx, cross_prod):
                try:
                    c_slice = getattr(slice_, c.local_name)[new_index]
                    if type(c_slice) is IndexedComponent_slice:
                        # This is just to make sure we do not have an
                        # empty slice.
                        temp_slice = c_slice.duplicate()
                        next(iter(temp_slice))
                    yield sliced_sets, c_slice
                except StopIteration:
                    # We have an empty slice for some reason, e.g.
                    # a coordinate of `new_index` from the cross
                    # product was skipped in the original component.
                    pass
                except KeyError:
                    # We are creating scalar components from a product of
                    # sets. Components may be undefined for certain indices.
                    # We want to simply skip that index and move on.
                    pass
        else:
            # `c` is indexed only by sets we would like to slice.
            # Slice the component if it is indexed so a future getattr
            # will be valid.
            try:
                if c.is_indexed():
                    c_slice = getattr(slice_, c.local_name)[...]
                    # Make sure this slice is not empty...
                    next(iter(c_slice.duplicate()))
                else:
                    c_slice = getattr(slice_, c.local_name)
                yield sliced_sets, c_slice
            except StopIteration:
                pass

    # We now descend into subblocks
    for sub in b.component_objects(Block, descend_into=False):
        subsets = list(sub.index_set().subsets())
        temp_idx = [get_slice_for_set(s) if s in sets else _NotAnIndex
                for s in subsets]
        new_sets = [s for s in subsets if s in sets]
        other_sets = [s for s in subsets if s not in sets]

        # For each set we are slicing, if the user specified an index, put it
        # here. Otherwise, slice the set and we will call next(iter(slice_))
        # once the full index is constructed.
        descend_index_sliced_sets = tuple(index_map[s] if s in index_map else 
                get_slice_for_set(s) for s in new_sets)

        # Extend stack with new matched indices.
        index_stack.extend(new_sets)

        if other_sets and sub.is_indexed():
            cross_prod = other_sets[0].cross(*other_sets[1:])

            for new_index in _fill_indices_from_product(temp_idx, cross_prod):
                try:
                    sub_slice = getattr(slice_, sub.local_name)[new_index]

                    # Now we need to pick a block data to descend into
                    if new_sets:
                        # We sliced some sets, and need to fill in any
                        # indices provided by the user

                        # `new_index` could be a scalar, for compatibility with
                        # `normalize_index.flatten==False`.
                        tupl_new_index = (new_index,) if type(new_index) \
                                is not tuple else new_index
                        # Extract the indices of "other sets":
                        incomplete_descend_index = list(
                                idx if subset not in sets else _NotAnIndex
                                for idx, subset in zip(tupl_new_index, subsets)
                                )
                        # Fill rest of the entries with specified indices for
                        # sliced sets:
                        descend_index = _fill_indices(incomplete_descend_index,
                                descend_index_sliced_sets)
                        if len(descend_index) == 1:
                            descend_index = descend_index[0]

                        descend_slice = sub[descend_index]
                        data = descend_slice if type(descend_slice) is not \
                            IndexedComponent_slice else next(iter(descend_slice))
                        # If the user has supplied enough indices that we can
                        # descend into a concrete component, we do so. Otherwise
                        # we use the user's indices, slice the rest, and advance
                        # the iterator.
                    else:
                        # All indices are specified
                        data = sub[new_index]
                    for st, v in generate_sliced_components(data, index_stack,
                            sub_slice, sets, ctype, index_map):
                        yield tuple(st), v
                except StopIteration:
                    # Empty slice due to "skipped" index in subblock.
                    pass
                #except KeyError:
                #    # Trying to access a concrete data object for a "skipped" index.
                #    # I have been unable to produce this behavior for blocks,
                #    # but it may be possible somehow.
                #    pass
        else:
            # Either `sub` is a simple component, or we are slicing
            # all of its sets. What is common here is that we don't need
            # to iterate over "other sets."
            try:
                if sub.is_indexed():
                    sub_slice = getattr(slice_, sub.local_name)[...]
                    # We have to get the block data object.
                    descend_slice = sub[descend_index_sliced_sets]
                    data = descend_slice if type(descend_slice) is not \
                        IndexedComponent_slice else next(iter(descend_slice))
                else:
                    # `sub` is a simple component
                    sub_slice = getattr(slice_, sub.local_name)
                    data = sub
                for st, v in generate_sliced_components(data, index_stack,
                        sub_slice, sets, ctype, index_map):
                    yield tuple(st), v
            except StopIteration:
                # We encountered an empty slice. This should be very rare.
                pass

        # pop the index sets of the block whose sub-components
        # we just finished iterating over.
        for _ in new_sets:
            index_stack.pop()

def flatten_components_along_sets(m, sets, ctype, indices=None):
    """
    This function iterates over components (recursively) contained
    in a block and partitions their data objects into components
    indexed only by the specified sets.

    Args:
        m : Block whose components (and their sub-components) will be
            partitioned
        sets : Possible indexing sets for the returned components
        ctype : Type of component to identify and partition
        indices : indices of sets to use when descending into subblocks

    Returns:
        tuple: The first entry is a list of tuples of Pyomo Sets. The
               second is a list of lists of components, each indexed by
               the corresponding sets in the first entry.
        
    """
    if indices is None:
        index_map = ComponentMap()
    elif type(indices) is ComponentMap:
        index_map = indices
    else:
        index_map = ComponentMap(zip(sets, indices))

    index_stack = []

    set_of_sets = ComponentSet(sets)
    # Using these two `OrderedDict`s is a workaround because I can't
    # reliably use tuples of components as keys in a `ComponentMap`.
    sets_dict = OrderedDict()
    comps_dict = OrderedDict()
    for index_sets, slice_ in generate_sliced_components(m, index_stack,
            m, set_of_sets, ctype, index_map):
        # Note that index_sets should always be a tuple, never a scalar.

        # TODO: Potentially re-order sets at this point.
        # In this way (time, space) would have the same key as (space, time).
        # They we'd have to somehow "swap indexing sets" when we create
        # the reference below.
        key = tuple(id(c) for c in index_sets)
        if key not in sets_dict:
            if len(key) == 0:
                sets_dict[key] = (UnindexedComponent_set,)
            else:
                sets_dict[key] = index_sets
        if key not in comps_dict:
            comps_dict[key] = []
        if len(key) == 0:
            comps_dict[key].append(slice_)
        else:
            # If the user wants to change these flags, they can access the
            # slice via the `referent` attribute of each reference component.
            slice_.attribute_errors_generate_exceptions = False
            slice_.key_errors_generate_exceptions = False
            comps_dict[key].append(Reference(slice_))

    # list-of-tuples of Sets:
    sets_list = list(sets for sets in sets_dict.values())
    # list-of-lists of components:
    comps_list = list(comps for comps in comps_dict.values())
    # E.g. we return: (
    #          [(time, space), (time,)],
    #          [[some_component, ...], [other, ...]],
    #      )                            ^ These components are indexed by time
    #            ^ These components are indexed by time and space
    return sets_list, comps_list

def flatten_dae_components(model, time, ctype, indices=None):
    target = ComponentSet((time,))
    sets_list, comps_list = flatten_components_along_sets(model, target, ctype,
            indices=indices)
    # Initialize these variables as, if no components of either category are
    # found, we expect to get an empty list.
    scalar_comps = []
    dae_comps = []
    for sets, comps in zip(sets_list, comps_list):
        if len(sets) == 1 and sets[0] is time:
            dae_comps = comps
        elif len(sets) == 0 or (len(sets) == 1 and 
                sets[0] is UnindexedComponent_set):
            scalar_comps = comps
        else:
            raise RuntimeError(
                "Invalid model for `flatten_dae_components`.\n"
                "This can happen if your model has components that are\n"
                "indexed by time (explicitly or implicitly) multiple times."
                )
    return scalar_comps, dae_comps



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
import inspect
from pyomo.common.deprecation import deprecation_warning
deprecation_warning(
    "The pyomo.core.base.plugin module is deprecated.  "
    "See pyomo.core.base.transformation for Transformation and "
    "TransformationFactory, pyomo.core.base.component for "
    "ModelComponentFactory and pyomo.scripting.interface for "
    "IPyomoScript* interfaces.",
    version='6.0', calling_frame=inspect.currentframe().f_back,
)

__all__ = ['pyomo_callback',
        'IPyomoExpression', 'ExpressionFactory', 'ExpressionRegistration',
        'IPyomoPresolver', 'IPyomoPresolveAction',
        'IParamRepresentation',
        'ParamRepresentationFactory',
        'IPyomoScriptPreprocess',
        'IPyomoScriptCreateModel',
        'IPyomoScriptCreateDataPortal',
        'IPyomoScriptModifyInstance',
        'IPyomoScriptPrintModel',
        'IPyomoScriptPrintInstance',
        'IPyomoScriptSaveInstance',
        'IPyomoScriptPrintResults',
        'IPyomoScriptSaveResults',
        'IPyomoScriptPostprocess',
        'ModelComponentFactory',
        'Transformation',
        'TransformationFactory',
        ]

from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.transformation import (
    Transformation, TransformationFactory, TransformationData,
    TransformationInfo, TransformationTimer
)

from pyomo.scripting.interface import (
    implements, Interface, Plugin, ExtensionPoint, DeprecatedInterface,
    pyomo_callback,
    IPyomoPresolver,
    IPyomoPresolveAction,
    IPyomoScriptPreprocess,
    IPyomoScriptCreateModel,
    IPyomoScriptCreateDataPortal,
    IPyomoScriptModifyInstance,
    IPyomoScriptPrintModel,
    IPyomoScriptPrintInstance,
    IPyomoScriptSaveInstance,
    IPyomoScriptPrintResults,
    IPyomoScriptSaveResults,
    IPyomoScriptPostprocess,
)


class IPyomoExpression(DeprecatedInterface):

    def type(self):
        """Return the type of expression"""

    def create(self, args):
        """Create an instance of this expression type"""


class IParamRepresentation(DeprecatedInterface):
    pass

